//
// Created by tim on 05.04.23.
//

#include <fstream>
#include "CTUApexConsumer.h"
#include "Util.h"
#include "llvm/Demangle/Demangle.h"
#include "FunctionFinderVisitor.h"

static bool generatedMain = false;

void CTUApexConsumer::outlineFunctions() {
    for (const auto &node: call_graph->getNodes()) {
        //check for each node in the graph, whether we found a corresponding node in the current ast
        if (auto md = node.second->checkAndGet<ExtractionMetadata>(); md.first && md.second->get_decl_node()) {
            ExtractionBlock eb;
            clang::SourceRange sourceRange;
            //The decl will refer to the instantiated template, but we need the contents of the original template specification
            if (md.second->get_decl_node()->isTemplateInstantiation()) {
                sourceRange = md.second->get_decl_node()->getPrimaryTemplate()->getSourceRange();
            } else {
                sourceRange = md.second->get_decl_node()->getSourceRange();
            }
            eb.begin = sourceRange.getBegin();
            eb.end = sourceRange.getEnd();
            //We only copy the contents if we are interested in them, but we do need an source-range bound extraction block for every function
            //To later filter for pragmas and macros
            if (md.second->isMarkedToExtract()) {
                eb.contents = clang::Lexer::getSourceText(clang::CharSourceRange::getTokenRange(sourceRange),
                                                          md.second->get_decl_node()->getASTContext().getSourceManager(),
                                                          clang::LangOptions(),
                                                          nullptr).str();
            } else {
                eb.contents = "";
            }
            funcs.push_back(eb);
        }
    }

    //the callgraph nodes are unsorted, we need to sort the extraction blocks afterward
    std::sort(funcs.begin(), funcs.end(),
              [](const ExtractionBlock &a, const ExtractionBlock &b) { return a.begin < b.begin; });
}

void CTUApexConsumer::outlineGlobals() {
    //Fixme: does this need to extract cross TU as well?

    std::unordered_set<clang::Decl *> availableGlobals;
    std::vector<ExtractionBlock> returnBlocks;
    for (const auto &node: call_graph->getNodes()) {
        if (auto md = node.second->checkAndGet<ExtractionMetadata>(); md.first && md.second->get_decl_node()) {
            assert(clang::isa<clang::FunctionDecl>(md.second->get_decl_node()) &&
                   "Extractionset should only contain function decls at this point");
            GlobalsMarker gMarker;
            auto globalsSet = gMarker.get_nodes_to_outline(
                    clang::cast<clang::FunctionDecl>(md.second->get_decl_node()));
            for (const auto &decl: globalsSet) {
                ExtractionBlock eb;
                eb.begin = decl->getBeginLoc();
                eb.end = decl->getEndLoc();
                clang::CharSourceRange charSourceRange({eb.begin, eb.end}, true);
                eb.contents = clang::Lexer::getSourceText(charSourceRange, decl->getASTContext().getSourceManager(),
                                                          clang::LangOptions(),
                                                          nullptr).str();
                globals.push_back(eb);
            }
        }
    }
    //as we find globals in the order they are used, but want them in the order they are declared we need to sort the globals
    std::sort(globals.begin(), globals.end(),
              [](const ExtractionBlock &a, const ExtractionBlock &b) { return a.begin < b.begin; });
}

void CTUApexConsumer::filterExtrationBlocks() {
    //Funcs has source-ranges for every function, but only contents for the relevant ones
    //We remove all entries in the includes, macros, and pragmas that are inside ANY function
    //We either get the chars by copying the function, then we do not need them
    //Or we do not extract the function, then we do not want to copy the includes, macros and pragmas.
    //We only want "out of function" includes, macros and pragmas
    for (const auto &func: funcs) {
        clang::SourceRange s = {func.begin, func.end};
        //Cast to void to ignore return-value of remove_if
        (void) std::remove_if(includes.begin(), includes.end(),
                              [&s](const ExtractionBlock &e) { return s.fullyContains({e.begin, e.end}); });
        (void) std::remove_if(macros.begin(), macros.end(),
                              [&s](const ExtractionBlock &e) { return s.fullyContains({e.begin, e.end}); });
        (void) std::remove_if(pragmas.begin(), pragmas.end(),
                              [&s](const ExtractionBlock &e) { return s.fullyContains({e.begin, e.end}); });
    }
}

ExtractionBlock CTUApexConsumer::getNextExtractionBlock() {
    //find the earliest block by starting source location
    //we need to do a potentially expensive copy of the full struct
    //as we will erase parts of the vectors later, potentially invalidating pointers in the process
    ExtractionBlock min;
    if (!includes.empty()) {
        min = includes.front();
    } else if (!funcs.empty()) {
        min = funcs.front();
    } else if (!globals.empty()) {
        min = globals.front();
    } else if (!macros.empty()) {
        min = macros.front();
    } else if (!pragmas.empty()) {
        min = pragmas.front();
    }

    while (!funcs.empty() && min.begin > funcs.front().begin) {
        min = funcs.front();
    }
    while (!globals.empty() && min.begin > globals.front().begin) {
        min = globals.front();
    }
    while (!macros.empty() && min.begin > macros.front().begin) {
        min = macros.front();
    }
    while (!pragmas.empty() && min.begin > pragmas.front().begin) {
        min = pragmas.front();
    }

    //if the earliest blocks end is after another block, we fully contain the other block
    //we don't need to merge that block anymore
    const auto &pred = [&min](const auto &elem) { return min.end >= elem.end; };
    includes.erase(std::remove_if(includes.begin(), includes.end(), pred), includes.end());
    funcs.erase(std::remove_if(funcs.begin(), funcs.end(), pred), funcs.end());
    globals.erase(std::remove_if(globals.begin(), globals.end(), pred), globals.end());
    macros.erase(std::remove_if(macros.begin(), macros.end(), pred), macros.end());
    pragmas.erase(std::remove_if(pragmas.begin(), pragmas.end(), pred), pragmas.end());
    return min;
}


void CTUApexConsumer::HandleTranslationUnit(clang::ASTContext &Context) {
    //Whole path with line and column
    std::string path = Context.getSourceManager().getLocForEndOfFile(
            Context.getSourceManager().getMainFileID()).printToString(Context.getSourceManager());
    //Only file name with line and column
    path = path.substr(path.find_last_of('/') + 1);
    //Only file name
    path = path.substr(0, path.find_first_of(':'));
    SPDLOG_INFO("Extracting from file: {}", path);

    const auto &translation_unit = Context.getTranslationUnitDecl();

    SPDLOG_INFO("Mapping functions to the call graph");

    FunctionFinderVisitor functionFinderVisitor(call_graph);
    functionFinderVisitor.TraverseTranslationUnitDecl(translation_unit);

    SPDLOG_INFO("Outlining Functions");
    outlineFunctions();

    SPDLOG_INFO("Outlining Globals");
    outlineGlobals();

    SPDLOG_INFO("Removing duplicate extraction blocks");
    filterExtrationBlocks();

    SPDLOG_INFO("Generate code skeleton");
    const auto &sm = Context.getSourceManager();
    std::string filename = sm.getFileEntryForID(sm.getMainFileID())->getName().str();
    filename.insert(filename.find_last_of('.') + 1, "mini.");
    std::ofstream outFile;
    outFile.open(filename);
    assert(outFile.is_open());

    while (!(includes.empty() && funcs.empty() && globals.empty() && macros.empty() && pragmas.empty())) {
        outFile << getNextExtractionBlock().contents << "\n";
    }

    SPDLOG_INFO("Generate wrapper calls");
    SPDLOG_CRITICAL("WRAPPER CALLS CAN NOT YET BE IMPLEMENTED");

    if (auto md = call_graph->getMain()->checkAndGet<ExtractionMetadata>(); md.first &&
                                                                            !md.second->isMarkedToExtract() &&
                                                                            !generatedMain) {
        //we have no main function and need to create or own
        outFile << "\nint main(){return 0;}\n";
        generatedMain = true;
    }


    // Outline the code we want
    //SPDLOG_DEBUG("Generating outline");
    //assert(output.empty());
    //ASTOutliner outliner(&context, dump_anal_res);
    //auto extracted_source = outliner.generate_outline(marked, only_wrapper);

    //if (!only_wrapper) {
    //    auto main_source = outliner.generate_main_function(marked);
    //    output = extracted_source + main_source;
    //} else {
    //    output = extracted_source;
    //}

    outFile.close();

    SPDLOG_INFO("Clean CG for next run");
    for (const auto &node: call_graph->getNodes()) {
        if (node.second->has<ExtractionMetadata>()) {
            node.second->get<ExtractionMetadata>()->set_decl_node(nullptr);
        }
    }
}




