//Wir wollen in der lage sein commandozeilen parameter zu verarbeiten
#include <clang/Tooling/CommonOptionsParser.h>
//Wir sind eine frontend action
#include <clang/Frontend/FrontendActions.h>
//Wir machen Tooling
#include "clang/Tooling/Tooling.h"
//Wir benutzen die Visitor strategy
//Wir müssen mit dem compiler interagieren können.
#include <clang/Frontend/CompilerInstance.h>
#include <spdlog/spdlog.h>
#include <iostream>
#include "CTUApexAction.h"
#include "CTUApexConsumer.h"

#include "io/VersionTwoMCGReader.h"
#include "io/VersionThreeMCGReader.h"

#include "Util.h"
#include "ExtractionMetadata.h"

#include "CommandLineArgs.h"

#include "llvm/Demangle/Demangle.h"

//Das sind convenience definitionen, dass wir nicht so viel tippen müssen
using namespace llvm;
using namespace clang;
using namespace clang::tooling;

template<typename T, class... Args>
std::unique_ptr<clang::tooling::FrontendActionFactory> argumentParsingFrontendActionFactory(Args... args) {

    class SimpleFrontendActionFactory : public clang::tooling::FrontendActionFactory {
    public:
        explicit SimpleFrontendActionFactory(Args... constructorArgs) {
            fn1 = [=]() { return std::make_unique<T>(constructorArgs...); };
        }

        std::unique_ptr<clang::FrontendAction> create() override {
            return fn1();
        }

    private:
        std::function<std::unique_ptr<T>()> fn1; // function
    };

    return std::unique_ptr<FrontendActionFactory>(new SimpleFrontendActionFactory(args...));
}

class CTUApexFrontendAction : clang::ASTFrontendAction {
public:
    CTUApexFrontendAction(metacg::Callgraph &cg) : cg(&cg) {}

    std::unique_ptr<clang::ASTConsumer> newASTConsumer() {
        return std::unique_ptr<clang::ASTConsumer>(new CTUApexConsumer(cg, includes, macros, pragmas));
    }


    std::unique_ptr<clang::ASTConsumer> CreateASTConsumer([[maybe_unused]] clang::CompilerInstance &compiler,
                                                          [[maybe_unused]] llvm::StringRef sr) {
        return std::unique_ptr<clang::ASTConsumer>(new CTUApexConsumer(cg, includes, macros, pragmas));
    }

private:
    std::vector<ExtractionBlock> includes;
    std::vector<ExtractionBlock> macros;
    std::vector<ExtractionBlock> pragmas;
    metacg::Callgraph *cg;
};


void markCalleesToExtract(const size_t nodeID, const metacg::Callgraph &cg) {
    for (const auto &call: cg.getCallees(nodeID)) {
        //This check is probably? necessary in case of recursive functions
        if(auto [h,v]=call->checkAndGet<ExtractionMetadata>(); !h || !v->isMarkedToExtract()){
            call->getOrCreateMD<ExtractionMetadata>()->setMarkedToExtract(true);
            markCalleesToExtract(call->getId(), cg);
        }
    }
}

void markFunctionsAlongCallpaths(metacg::Callgraph &cg) {
    SPDLOG_INFO("Compute set of functions to be extracted");
    for (const auto &node: cg.getNodes()) {
        if (auto md = node.second->checkAndGet<ExtractionMetadata>();
                md.first && md.second->isMarkedToExtract()) {
            markCalleesToExtract(node.first, cg);
        }
    }
}


std::vector<std::string> clangIncludeBasedArgumentAdjustor(const CommandLineArguments &cla, StringRef sr) {
    //call the compiler specified in cla[0] with -S and -v like so
    std::string command = "clang";
    //command.append(" -S -v");
    command.append("  -W,p -v -fsyntax-only");
    for (int i = 1; i < cla.size(); i++) {
        command.append(" " + cla[i]);
    }
    command.append(" 2>&1");
    std::unique_ptr<FILE, decltype(&pclose)> pipe(popen(command.c_str(), "r"), pclose);
    //then read the output of the command like so:
    std::string txt;
    std::array<char, 128> buffer{};
    while (fgets(buffer.data(), buffer.size(), pipe.get()) != nullptr) {
        txt += buffer.data();
    }
    //search for "include search starts here" string and add all the paths with -I to the tool invocation
    txt = txt.substr(txt.rfind("search starts here:") + 20,
                     txt.rfind("End of search list.") - txt.rfind("search starts here:") - 20);
    std::vector<std::string> ret;
    ret.assign(cla.begin(), cla.end());
    std::string::size_type pos;
    std::string::size_type prev = 0;
    while ((pos = txt.find('\n', prev)) != std::string::npos) {
        auto t=std::prev(ret.end(),3);
        ret.insert(t,{"-I" + txt.substr(prev, pos - prev)});
        prev = pos + 2;
    }
    return ret;
}

std::vector<std::string> clangResourceDirBasedAdjuster(const CommandLineArguments &cla, StringRef) {
    //call the compiler specified in cla[0] with the job to print the resource directory for compiler specific includes
    std::string command = "clang --print-resource-dir 2>1&";
    std::unique_ptr<FILE, decltype(&pclose)> pipe(popen(command.c_str(), "r"), pclose);
    //then read the output of the command like so:
    std::string txt;
    std::array<char, 128> buffer{};
    while (fgets(buffer.data(), buffer.size(), pipe.get()) != nullptr) {
        txt += buffer.data();
    }
    std::vector<std::string> ret;
    ret.assign(cla.begin(), cla.end());
    trim(txt);
    //we do need to strip the newline char at the end
    txt.pop_back();
    //the compiler jobs last entries are the file, the denotion for positional arguments and the syntax only flag
    //additional search paths need to be inserted before these 3
    auto t=std::prev(ret.end(),3);
    ret.insert(t,{"-I" + txt + "/include"});
    return ret;
}

void printMarkedFunctions(const metacg::Callgraph &cg) {
    std::cout << "Extracting:\n";
    size_t size=500;
    char* a = static_cast<char *>(malloc(size * sizeof(char)));
    int status;
    size_t count=0;
    for (const auto &[_, node]: cg.getNodes()) {
        if (node->has<ExtractionMetadata>()) {
            count++;
            if(llvm::itaniumDemangle(node->getFunctionName().c_str(),a,&size,&status);status==llvm::demangle_success){
                if(llvm::StringRef b =a;!b.startswith("std::")){
                    //std::cout << "Node: " << demangle(node->getFunctionName()) << "\n";
                }
            }
        }
    }
    free(a);
    std::cout<<"Marked: "<<count<<" functions\n";
}


int main(int argc, const char **argv) {

    auto ExpectedParser = CommonOptionsParser::create(argc, argv, CTUApex);

    if (!ExpectedParser) {
        // Fail gracefully for unsupported options.
        llvm::errs() << ExpectedParser.takeError();
        return 1;
    }

    switch (LoggingLevel) {
        case LogLevel::Trace:
            spdlog::set_level(spdlog::level::trace);
            break;
        case LogLevel::Debug:
            spdlog::set_level(spdlog::level::debug);
            break;
        case LogLevel::Info:
            spdlog::set_level(spdlog::level::info);
            break;
        case LogLevel::Warning:
            spdlog::set_level(spdlog::level::warn);
            break;
        case LogLevel::Error:
            spdlog::set_level(spdlog::level::err);
            break;
        case LogLevel::Off:
            spdlog::set_level(spdlog::level::off);
            break;
    }

    CommonOptionsParser &OptionsParser = ExpectedParser.get();


    ClangTool tool(OptionsParser.getCompilations(), OptionsParser.getSourcePathList());
    tool.appendArgumentsAdjuster(clangResourceDirBasedAdjuster);

    metacg::io::FileSource fs(cg_file.getValue());

    std::unique_ptr<metacg::io::MetaCGReader> mcgReader;

    if (fs.get().at("_MetaCG").at("version") == "3.0") {
        mcgReader = std::make_unique<metacg::io::VersionThreeMetaCGReader>(fs);
    } else if (fs.get().at("_MetaCG").at("version") == "2.0") {
        mcgReader = std::make_unique<metacg::io::VersionTwoMetaCGReader>(fs);
    } else {
        assert(false && "File corrupt or format not specified");
    }

    auto &mcgManager = metacg::graph::MCGManager::get();

    //As the mcgReader spits out a warning for every occurrence of a missing metadata,
    //we silence all warnings for the duration of the read
    const auto lv = spdlog::get_level();
    spdlog::set_level(spdlog::level::off);
    mcgManager.addToManagedGraphs("activeGraph", mcgReader->read());
    spdlog::set_level(lv);

    auto &cg = *mcgManager.getCallgraph();

    //Add extraction point gotten via the CLI
    for (const auto &extractionFunc: extractionPoints) {
        if (cg.hasNode(extractionFunc)) {
            auto *md = new ExtractionMetadata();
            md->setMarkedToExtract(true);
            cg.getNode(extractionFunc)->addMetaData(md);
        } else {
            SPDLOG_ERROR("The given function: " + extractionFunc + " does not exist in the given callgraph");
        }
    }



    markFunctionsAlongCallpaths(cg);

    //printMarkedFunctions(cg);

    tool.run(argumentParsingFrontendActionFactory<CTUApexAction>(mcgManager.getCallgraph()).get()) ;
    return 0;
}